'''Gather corresponding non-coding sequences from genome groups'''

import os                     # for directory operations
from Bio import SeqIO         # for parsing Genbank files
import ncds_config as config  # global configuration parameters

def main():
    '''Gather corresponding non-coding sequences from genome groups'''
    _say('NCDS: Non-CoDing Sequences')
    _say('')
    _check_dir_is_empty(config.output_dir)
    _say("Reading families list from '{0}'.".format(config.families_file))
    families = read_families(config.families_file)
    _say('Checking for the presence of all needed genome files.')
    check_genomes_presence(families, config.genomes_dir)
    _say('')
    for family_name, family in families:
        _say("Processing family '{0}'.".format(family_name))
        family_genomes = {}  # maps genome names to genome (SeqRecord) objects
        family_ncds = {}     # maps genome names to genome NCDS dict
        for genome_name in family:
            _say("Genome '{0}':".format(genome_name), False)
            _say('parsing...', False)
            genome = read_genome(genome_name, config.genomes_dir)
            ### temp: to filter specific taxa:
            #if 'Viridiplantae' in genome.annotations['taxonomy']:
            #    print '   ', genome_name, ' #', genome.annotations['organism']
            #continue
            family_genomes[genome_name] = genome
            _say('features...', False)
            genome_features = prepare_genome_features(genome)
            _say('NCDS list...', False)
            family_ncds[genome_name] = prepare_genome_ncds(genome_features)
            _say('done.')
        _say('Gathering NCDS groups...', False)
        ncds_groups = gather_ncds_groups(family_ncds)
        _say('{0} groups,'.format(len(ncds_groups)), False)
        ncds_groups = {gid: group for gid, group in ncds_groups.iteritems()
                       if len(group) >= config.min_ncds_group_size}
        _say('{0} large groups.'.format(len(ncds_groups)))
        if config.remove_duplicates:
            _say('Removing duplicates...', False)
            remove_duplicates(ncds_groups, family_genomes)
            _say('done.')
        _say('Writing NCDS groups to FASTA files...     ', False)
        progress_indicator = lambda n, m: \
            _say('\b'*5 + '{0:3d}%'.format(not m and 100 or 100*n/m), False)
        output_ncds_groups(ncds_groups, family_ncds, family_genomes,
                           family_name, config.output_dir,
                           progress_indicator)
        _say('done.')
        _say("Finished processing family '{0}'.".format(family_name))
        _say('')
    _say('All done.')

def _say(something, new_line=True):
    '''Send a message to... stdout'''
    if new_line:
        print something
    else:
        print something,

def _check_dir_is_empty(name):
    '''Check that directory with given name is empty'''
    try:
        if len(os.listdir(name)):
            exit("Stop. Output directory '{0}' is not empty. ".format(name) +
                 "Please clear it manually to prevent accidental data loss.")
    except OSError as (errno, errstr):
        if errno != 3:  # not exists
            exit("Bad output directory '{name}':\n{errstr}".format(**vars()))

def read_families(filename):
    '''Read families list from file'''
    def error(message, filename, ln, line):
        exit(("Error in families file '{filename}', line {ln}: {message}\n" +
              "{line}").format(**vars()))
    family_names = []
    family_members = {}

    try:
        with file(filename) as infile:
            lines = infile.readlines()
    except IOError as (n, e):
        exit("Can't read families file '{filename}:\n{e}'".format(**vars()))
    for ln, line in enumerate(lines, 1):
        indented = line[0] in (' ', '\t')
        if '#' in line:
            line = line[:line.index('#')]
        line = line.strip()
        if not line:
            continue
        if not indented:  # it's a family name line
            if not line.endswith(':'):
                error("family name line should end with ':'",
                      filename, ln, line)
            if family_names and not family_names[-1] in family_members:
                error('family name should be followed by a members list',
                      filename, ln, line)
            family_names.append(line[:-1].strip())
        else:             # it's a family members list line
            if not family_names:
                error('members list should follow a family name',
                      filename, ln, line)
            if line.endswith(','):
                line = line[:-1]
            members = [member.strip() for member in line.split(',')]
            family_members.setdefault(family_names[-1], []).extend(members)
    families = [(name, family_members[name]) for name in family_names]
    return families

def check_genomes_presence(families, genomes_dir):
    '''Check all genome files are present'''
    for genome_name in sum(map(lambda x: x[1], families), []):
        file_name = os.path.join(genomes_dir, genome_name + '.gbk')
        if not os.path.exists(file_name):
            exit("Genome file is absent: '{0}'.".format(file_name))

import warnings  # to suppress parser warning

def read_genome(genome_name, genomes_dir):
    '''Read genome with given name from .gbk file in given directory'''
    infile_name = os.path.join(genomes_dir, genome_name + '.gbk')
    try:
        with file(infile_name) as infile:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "^New parser didn't like ")
                genome = SeqIO.read(infile, 'genbank')
    except IOError as (n, e):
        exit("I/O error while reading genome file '{0}':\n{1}.".\
             format(infile_name, e))
    return genome

def prepare_genome_features(genome):
    '''Prepare a dictionary of the given genome features.

    Records in the returned dictionary have the following structure:
      (left_coord, right_coord): Feature_object_having_these_coordinates
    '''
    # gather features info
    features = get_features_dict(genome)
    # join features with the same coordinates (exons, etc.)
    join_features(features)
    return features

def prepare_genome_ncds(features):
    '''Prepare a dictionary of NCDS for the given genome features.

    Records in the returned dictionary have the following structure:
      (left_coord, right_coord): (left_neighbours_list, right_neighbours_list)
    The items in each neighbours list are Feature objects.
    '''
    # gather NCDS positions
    ncds_coordinates = get_ncds_coordinates(features)
    # gather left and right neighbours for each NCDS
    ncds = find_neighbours(ncds_coordinates, features)
    return ncds

def gather_ncds_groups(ncds):
    '''Gather NCDS in groups by NCDS group ids.

    Records in the returned dictionary have the following structure:
      ncds_gid: {genome1_name: genome1_ncds_group_members, ...}
    where each ncds_group_members is a set of tuples:
      (left_coordinate, right_coordinate, changed_order_flag)
    '''
    ncds_groups = {}  # maps NCDS group ids to NCDS groups
    for genome_name, ncds_dict in ncds.iteritems():
        for (left, right), (left_genes, right_genes) in ncds_dict.iteritems():
            if not left_genes or not right_genes:
                continue
            for left_gene in left_genes:
                if not left_gene.name:
                    continue
                for right_gene in right_genes:
                    if not right_gene.name:
                        continue
                    ncds_gid = (left_gene.name4id, right_gene.name4id)
                    order_changed = False
                    if ncds_gid[0] > ncds_gid[1]:
                        ncds_gid = (ncds_gid[1], ncds_gid[0])
                        order_changed = True
                    ncds_groups.setdefault(ncds_gid, {}). \
                                setdefault(genome_name, set()). \
                                add((left, right, order_changed))
    return ncds_groups

def remove_duplicates(ncds_groups, family_genomes):
    '''Remove exact duplicates of NCDS within each group->genome'''
    for n, (gid, group) in enumerate(ncds_groups.items(), 1):
        for genome_name, ncds in sorted(group.items()):
            seqs = set()
            for left, right, is_reversed in sorted(ncds):
                seq = family_genomes[genome_name][left:right].seq
                if is_reversed:
                    seq = seq.reverse_complement()
                seq = str(seq)
                if seq in seqs:
                    ncds_groups[gid][genome_name].remove((left, right,
                                                          is_reversed))
                else:
                    seqs.add(seq)

def output_ncds_groups(ncds_groups, family_ncds, family_genomes, family_name,
                       output_dir, progress_indicator):
    '''Send each NCDS group to it's own FASTA file'''
    outdir_name = os.path.join(output_dir, str(family_name))
    try:
        os.makedirs(outdir_name)
    except OSError as (n, e):
        exit("Can't create directory '{0}':\n{1}.".format(outdir_name, e))
    with file(os.path.join(outdir_name, 'info.txt'), 'w') as info_file:
        info_file.write("Genomes summary for '{0}'\n\n".format(family_name))
        for gn, (gid, genome) in enumerate(sorted(family_genomes.items()), 1):
            info_file.write('{0}. "{1}"\n'.format(gn, gid))
            info_file.write(genome_info(genome))
            info_file.write('\n')
    files_count = len(ncds_groups)
    progress_indicator(0, files_count)
    for n, (gid, group) in enumerate(ncds_groups.items(), 1):
        file_name = '{0:02d}'.format(len(group)) + '.' + '..'.join(gid)
        file_name = _replace_illegal_filename_characters(file_name)
        file_name = os.path.join(outdir_name, file_name + '.fasta')
        with file(file_name, 'w') as outfile:
            for genome_name, ncds in sorted(group.items()):
                genome = family_genomes[genome_name]
                g_ncds = family_ncds[genome_name]
                #seqs = set()  # unique sequences
                for left, right, is_reversed in sorted(ncds):
                    left_neighbours, right_neighbours = g_ncds[left, right]
                    coordinates = '{0}..{1}'.format(left+1, right)
                    seq = genome[left:right].seq
                    if is_reversed:
                        coordinates = 'complement({0})'.format(coordinates)
                        seq = seq.reverse_complement()
                    #if str(seq) in seqs:
                    #    continue  # skip the duplicate sequence
                    #seqs.add(str(seq))
                    outfile.write('>{0}:{1}'.format(genome_name, coordinates))
                    for left_neighbour in left_neighbours:
                        outfile.write(' left:' + str(left_neighbour))
                    for right_neighbour in right_neighbours:
                        outfile.write(' right:' + str(right_neighbour))
                    outfile.write('\n' + str(seq) + '\n')
        progress_indicator(n, files_count)


def _replace_illegal_filename_characters(s, c=b'~'):
    '''Replace illegal characters in filename with given character'''
    from string import maketrans
    BAD_CHARS = b"\"':\\|/?*"
    return s.translate(maketrans(BAD_CHARS, c*len(BAD_CHARS)))

def get_features_dict(genome):
    '''Gather genome features in {coordinates: [Feature list]} dict'''
    features = {}
    for f in genome.features:
        # not interested in some features here, especially introns!
        if f.type not in ('CDS', 'tRNA', 'rRNA', 'exon'):
            continue
        gene_name = f.qualifiers.get('gene', [''])[0]
        # ignore no-name genes if neaded
        if config.ignore_nameless_genes and not gene_name:
            continue
        sub_features = f.sub_features
        if not sub_features:
            sub_features = [f]
        for n, sf in enumerate(sub_features):
            # not interested in sub-features located in external places
            if sf.ref:
                continue
            exon_number = f.qualifiers.get('number', [''])[0]
            # try to guess exon number...
            if not exon_number and len(sub_features) > 1 and \
               'trans_splicing' not in f.qualifiers:
                exon_number = str(f.strand == 1 and
                                  n + 1 or len(sub_features) - n)
            ft = Feature(start=sf.location.nofuzzy_start,
                         end=sf.location.nofuzzy_end,
                         strand=sf.strand,
                         type_=f.type,
                         name=gene_name,
                         exon=exon_number)
            features.setdefault((ft.start, ft.end), []).append(ft)
    return features

def join_features(features):
    '''Join features having the same coordinates'''
    for (start, end), feats in features.items():
        feature = feats.pop()
        for f in feats:
            # check the two genes have the same direction
            if f.strand != feature.strand:
                f.strand = 0
                #print ' * Ahtung! Two genes have the same coordinates ' + \
                #      '{0}..{1} but different dirrections!'. \
                #      format(start + 1, end)
            # choose exon number (if they differ)
            if f.exon != feature.exon:
                if not f.exon or not feature.exon:
                    feature.exon += f.exon
                elif config.reject_doubtful_genes:
                    # we better not to use this to minimize error probability
                    feature.name = ''
                    break
                else:
                    print ' * Warning: exon numbers do not match:', \
                          '{0} != {1}.\n   {2}\n   {3}'. \
                          format(f.exon, feature.exon, f, feature)
                    if f.type == 'exon' and feature.type != 'exon':
                        feature.exon = f.exon
            # set feature type
            if f.type != 'exon':
                if feature.type == 'exon':
                    feature.type = f.type
                else:
                    feature.type += '+' + f.type
            # sett feature name
            if f.name != feature.name:
                if f.name and feature.name:
                    feature.name += '+' + f.name
                else:
                    feature.name += f.name
        features[start, end] = feature

def get_ncds_coordinates(features):
    '''Return genome NCDS coordinates set'''
    ncds_coordinates = set()
    cur_max_end = None
    for start, end in sorted(features.keys()):
        if start > end:
            exit('Panic: the start is greater than the end!')
        if cur_max_end and start > cur_max_end:
            ncds_coordinates.add((cur_max_end, start))
        cur_max_end = max(cur_max_end, end)
    return ncds_coordinates

def find_neighbours(ncds_coordinates, features):
    '''Find left and right neighbours for each NCDS'''
    index = dict()
    for coordinates, feature in features.iteritems():
        if feature.name:
            for c in coordinates:
                index.setdefault(c, []).append(feature)
    ncds = {}
    for left, right in sorted(ncds_coordinates):
        ncds[left, right] = (index.get(left, []), index.get(right, []))
    return ncds

def genome_info(genome):
    '''Prepare genome info summary'''
    info = ''
    info += 'Accession: {0}\n'.format(genome.name)
    info += 'Definition: {0}\n'.format(genome.description)
    info += 'Source: {0}\n'.format(genome.annotations['source'])
    info += 'Organism: {0}\n'.format(genome.annotations['organism'])
    info += 'Taxonomy: {0}\n'.format('/'.join(genome.annotations['taxonomy']))
    info += 'Date: {0}\n'.format(genome.annotations['date'])
    if 'comment' in genome.annotations:
        info += 'Comment:\n'
        info += '    ' + genome.annotations['comment'].replace('\n', '\n    ')
    info += '\n'
    return info

class Feature(object):
    '''Contains essencial information about genome feature (e.g. gene)'''
    def __init__(self, start, end, strand, type_, name, exon):
        self.start = start
        self.end = end
        self.strand = strand
        self.type = type_
        self.name = name
        self.exon = exon
    def __repr__(self):
        return 'Feature' + repr((self.start, self.end, self.strand,
                                 self.type, self.name, self.exon))
    def __str__(self):
        text = self.type + '<' + self.name + '>'
        if self.exon:
            text += '#' + self.exon
        text += '@' + self.coordinates_str
        return text

    @property
    def coordinates_str(self):
        coordinates = '{0}..{1}'.format(self.start + 1, self.end)
        if self.strand == -1:
            coordinates = 'complement({0})'.format(coordinates)
        elif not self.strand:
            coordinates = 'unknown({0})'.format(coordinates)
        return coordinates

    @property
    def name4id(self):
        name, exon = self.name, self.exon
        # remove '-XXX' at the end, where X is in 'AUGC'
        if len(name) > 4 and name[-4] == '-' and \
           not set(name[-3:]) - set('AUGCaugc'):
            name = name[:-4]
        # remove '(XXX)' at the end, where X is in 'AUGC'
        if len(name) > 5 and name[-5:-4] + name[-1] == '()' and \
           not set(name[-4:-1]) - set('AUGCaugc'):
            name = name[:-5].strip()
        # ... like 'trnHGUG'
        if len(name) == 7 and name[:3] == 'trn' and \
           not set(name[-3:]) - set('AUGCaugc'):
            name = name[:4]
        if exon:
            name += '#' + exon
        return name

if __name__ == '__main__':
    main()
